/******************************************************************/
/**@file opendev.c:
 * Desctiption: Humusoft data acquisition Linux library
                implementation.                                   *
 * Dependency: Linux                                              *
 *                Copyright 2006-2019 Jaroslav Fojtik             *
 ******************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <dirent.h>

#include "hudaqlib.h"
#include "hudaq_internal.h"

DWORD *SearchForDevice(const char *DevName, int *DeviceCount);


/** This function extracts one uint16 value from the file specified. */
static int sysfs_get_value(const char *FileName)
{
FILE *F;
int RetVal;

  F=fopen(FileName,"rb");
  if(F==NULL) return -1;

  fscanf(F,"%X",&RetVal);
  fclose(F);
return RetVal;
}


static int DeviceIDFound(const DWORD *DeviceIDs, int IDsCount, DWORD DeviceID)
{
  if(DeviceIDs==NULL || IDsCount<=0) return 0;
  while(IDsCount-->0)
    {
    if(DeviceID == DeviceIDs[IDsCount])
      {
      //printf("Try to match %d physical: %X listed: %X\n",IDsCount, DeviceID, DeviceIDs[IDsCount]);
      return 1;
      }
    }
  return 0;
}


////////////////////////////////////////////////////////////////////

static DWORD ScanSys(DWORD *DeviceIDs, int IDsCount, int deviceorder, DriverRecord *Rec)
{
DIR *dir;
FILE *F;
char ch;
struct dirent *entry;
unsigned vendor_id;
unsigned device_id;
char FileName[1024];
const char *SysDir="/sys/bus/pci/devices";
unsigned long long start, end, size;
DWORD Resources[14];
int i;
DWORD RetVal = 0;

  if(deviceorder <= 0) return  0xFFFFFF00 | 1;
  if(Rec==NULL) return  0xFFFFFF00 | 2;
  if(DeviceIDs==NULL || IDsCount<=0) return  0xFFFFFF00 | 3;

  Rec->NumPhysMemResources = Rec->Resources.NumIOResources = 0;

  sprintf(FileName,"%s",SysDir);
  dir = opendir(FileName);
  if(!dir)
    {
    //fprintf(stderr,"Cannot open /sys/bus/pci/devices");
    return  0xFFFFFF00 | 4;
    }

  while((entry = readdir(dir)))
    {
    if(entry->d_name[0]=='.') continue;

    memset(Resources, 0, sizeof(Resources));

    sprintf(FileName,"%s/%s/vendor",SysDir,entry->d_name);
    vendor_id = sysfs_get_value(FileName);
    sprintf(FileName,"%s/%s/device",SysDir,entry->d_name);
    device_id = sysfs_get_value(FileName);

    if(DeviceIDFound(DeviceIDs, IDsCount, (device_id|(vendor_id<<16))))
      {
      Rec->Resources.VendorID = vendor_id;
      Rec->Resources.DeviceID = device_id;

      //fprintf(stderr,"\nDEVICE %X HAS BEEN FOUND!", (device_id|(vendor_id<<16)));

      sprintf(FileName,"%s/%s/resource",SysDir,entry->d_name);
      F = fopen(FileName,"rb");
      if(F)
        {
        for(i=0; i<7; i++)
          {
          if(fscanf(F,"%llx %llx", &start, &end) != 2)
	    {
	    fprintf(stderr,"\nSyntax error in %s", FileName);
	    break;
	    }
	  do
	    {  
	    ch=fgetc(F);
            if(feof(F)) break;
            } while(ch!='\r' && ch!='\n' && ch!=0);  
	  
          if(start)
	    size = end - start + 1;
          else
	    size = 0;
	  
	  Resources[i]=start;
	  Resources[i+7]=size;
          }
        fclose(F);
        }

      if(--deviceorder<1)
        {
        for(i=0; i<7; i++)
          {
          if(Resources[i]!=0)
            {
            if(Resources[i] > 0xFFFF)
               {
               Rec->PhysMemResources[Rec->NumPhysMemResources].Base=Resources[i];
               Rec->PhysMemResources[Rec->NumPhysMemResources].Length=Resources[i+7];
               Rec->NumPhysMemResources++;
               }
            else
               {
               Rec->Resources.IOResources[Rec->Resources.NumIOResources].Base=Resources[i];
               Rec->Resources.IOResources[Rec->Resources.NumIOResources].Length=Resources[i+7];
               Rec->Resources.NumIOResources++;
               }
            }
          }
        RetVal = device_id | (vendor_id<<16);  //device has been found!!
        break;
        }
      }

//  printf("\n%X %X %s\n", vendor_id, device_id, entry->d_name);
    }

  closedir(dir);
  return RetVal;
}


static DWORD ScanProc(DWORD *DeviceIDs, int IDsCount, int deviceorder, DriverRecord *Rec)
{
FILE *devices;
DWORD DevTblID;
DWORD DevPos, DevInt;
DWORD Resources[14];
int i;
char ch;
DWORD RetVal = 0;

  if(deviceorder <= 0) return  0xFFFFFF00 | 1;
  if(Rec==NULL) return  0xFFFFFF00 | 2;
  if(DeviceIDs==NULL || IDsCount<=0) return  0xFFFFFF00 | 3;

  Rec->NumPhysMemResources = Rec->Resources.NumIOResources = 0;

  //devices=fopen("devices","rb");
  devices=fopen("/proc/bus/pci/devices","rb");
  if(devices==NULL) return  0xFFFFFF00 | 4;

    /* loop over devices */
  while (!feof(devices))
    {
    if(fscanf(devices,"%X %X %X", &DevPos, &DevTblID, & DevInt)!=3)
      return  0xFFFFFF00 | 5;

    memset(Resources, 0, sizeof(Resources));
    for(i=0; i<14; i++)
      {
      if(fscanf(devices,"%X", &Resources[i])!=1)
        {
	if(i==6 || i==7) break;  //correct
	break; //strange
	}
      }
    do {
       ch = fgetc(devices);
       if(feof(devices)) break;
       } while(ch!='\r' && ch!='\n' && ch!=0);

    if(DeviceIDFound(DeviceIDs, IDsCount, DevTblID))
      {
      Rec->Resources.VendorID = DevTblID >> 16;
      Rec->Resources.DeviceID = DevTblID & 0xFFFF;

      if(--deviceorder<1)
        {
        for(i=0; i<7; i++)
          {
          if(Resources[i]!=0)
            {
            if(Resources[i] & 1)
               {
	       Rec->Resources.IOResources[Rec->Resources.NumIOResources].Base = Resources[i] & ~1;
               Rec->Resources.IOResources[Rec->Resources.NumIOResources].Length=Resources[i+7];
               Rec->Resources.NumIOResources++;
               }
            else
               {
	       Rec->PhysMemResources[Rec->NumPhysMemResources].Base=Resources[i];
               Rec->PhysMemResources[Rec->NumPhysMemResources].Length=Resources[i+7];
               Rec->NumPhysMemResources++;
               }
            }
          }
        RetVal = DevTblID;    // Device has been found!!
        break;
        }
      }
    }

  fclose(devices);
return(RetVal);
}


/*static int ScanUIO(DWORD *DeviceIDs, int IDsCount, int deviceorder, DriverRecord *Rec)
{
DIR *dir_uio, *dir_memmaps;
FILE *F;
char ch;
struct dirent *direntry_uio, *direntry_memmaps;
unsigned vendor_id;
unsigned device_id;
char FileName[1024];
const char *SysDir="/sys/class/uio/";
unsigned long long start, end, size;
unsigned int index;
DWORD Resources[7 * 3];
int i;
int j = 0;
int RetVal = 0;

  if(deviceorder <= 0)	return -1;
  if(Rec == NULL) return -2;

  Rec->NumPhysMemResources = Rec->Resources.NumIOResources = 0;

  sprintf(FileName, "%s", SysDir);
  dir_uio = opendir(FileName);
  if(!dir_uio)
  {
    perror("opendir()");
    return -3;
  }

  while(direntry_uio = readdir(dir_uio))
  {
    if((strcmp((direntry_uio->d_name), ".") == 0) || 
		(strcmp((direntry_uio->d_name), "..") == 0))
    {
	continue;
    }

  sprintf(FileName, "%s/%s/device/vendor", SysDir, direntry_uio->d_name);
		vendor_id = sysfs_get_value(FileName);
		sprintf(FileName, "%s/%s/device/device", SysDir, direntry_uio->d_name);
		device_id = sysfs_get_value(FileName);


		//Fixme search also through iomaps; add into the same Resources array	
  memset(Resources, 0, sizeof(Resources)); 
  sprintf(FileName, "%s/%s/maps", SysDir, direntry_uio->d_name);
  dir_memmaps = opendir(FileName);
  if(!dir_memmaps) 
  {
    perror("opendir()");
    printf("%s\n", FileName);
    return -3;
  }
  while (direntry_memmaps = readdir(dir_memmaps)) 
  {
    if((strcmp(direntry_memmaps->d_name, ".") == 0) || (strcmp(direntry_memmaps->d_name, "..") == 0))
				continue;

    //printf("maps: %s\n", direntry_memmaps->d_name);
    sprintf(FileName, "%s/%s/maps/%s/addr", SysDir, direntry_uio->d_name, direntry_memmaps->d_name);
    F = fopen(FileName, "rb");
    if(F)
    {
      if(fscanf(F, "%llx", &start) != 1)
      {
	fprintf(stderr, "Syntax error in %s", FileName);
	break;
      }
      Resources[j] = start;
      fclose(F);
    }

    sprintf(FileName, "%s/%s/maps/%s/size", SysDir, direntry_uio->d_name, direntry_memmaps->d_name);
    F = fopen(FileName, "rb");
    if(F)
    {
      if(fscanf(F, "%llx", &size) != 1)
      {
	fprintf(stderr, "Syntax error in %s", FileName);
	break;
      }
      Resources[j + 7] = size;
		
      fclose(F);
    }

    index = atoi((direntry_memmaps->d_name) + 3); // "+3", because we have to skip over the word "map"
    Resources[j + 7 + 7] = index;
    j ++;
  }
		
  closedir(dir_memmaps);    

  if(DeviceIDFound(DeviceIDs, IDsCount, device_id + (vendor_id << 16)))
  {
    //printf("DEVICE HAS BEEN FOUND!\n");
    if(--deviceorder < 1)
    {
      for(i = 0; i < 7; i++)
      {
	if(Resources[i] != 0)
	{
          if(Resources[i] > 0xFFFF)
	  {
	    Rec->PhysMemResources[Rec->NumPhysMemResources].Base = Resources[i];
	    Rec->PhysMemResources[Rec->NumPhysMemResources].Length = Resources[i + 7];
	    Rec->PhysMemResources[Rec->NumPhysMemResources].UioMappingIndex	= 
	    Resources[i + 7 + 7];
	    Rec->NumPhysMemResources ++;
	  }
	  else
	  {
	    Rec->Resources.IOResources[Rec->Resources.NumIOResources].Base = Resources[i];
	    Rec->Resources.IOResources[Rec->Resources.NumIOResources].Length = Resources[i + 7];
	    Rec->Resources.IOResources[Rec->Resources.NumIOResources].UioMappingIndex =
								 Resources[i + 7 + 7];
            Rec->Resources.NumIOResources ++;
	  }
	}
      }

      Rec->UioDevNum = atoi((direntry_uio->d_name) + 3); //number of uio device

      RetVal = 1;    //device has been found!!
				break;
    }
  }

  //printf("%X %X %s\n", vendor_id, device_id, direntry_uio->d_name);
  }

  closedir(dir_uio);
  return RetVal;
} */


/*****************************************************************************
;*
;*              OpenDeviceHandle
;*              opens device handle by device name and order
;*
;*              Input:  device name
;*                      device order number
;*              Output: positive number when success; 0 or >0xFFFFFF00 when fails.
;*
;****************************************************************************/
unsigned OpenDeviceHandle(const char *devicename, int deviceorder, DriverRecord *Rec)
{
DWORD *DeviceIdList;
unsigned RetVal;
int DeviceMatch;

  if(deviceorder <= 0) return 0xFFFFFF00 | 1;
  if(Rec==NULL) return  0xFFFFFF00 | 2;

  DeviceIdList = SearchForDevice(devicename,&DeviceMatch);
  if(DeviceIdList==NULL || DeviceMatch==0)
	return 0xFFFFFF00 | 3;

//  fprintf(stderr,"\n%s deviceID = %X", devicename, DeviceMatch);

  RetVal = ScanSys(DeviceIdList, DeviceMatch, deviceorder, Rec);
//  fprintf(stderr,"\nScansys result %X", RetVal);
  if(RetVal==0 || RetVal>0xFFFFFF00)
    RetVal = ScanProc(DeviceIdList, DeviceMatch, deviceorder, Rec);

  free(DeviceIdList); DeviceIdList=NULL;
return(RetVal);
}

